unit ExLists;

interface

uses Windows, Classes, SysUtils, Contnrs, TypInfo, Graphics;

type
  TNotifyType = (ntLoadComplete, ntSaveComplete);
  
  //  .     RTTI 
  //       Published 
  TRttiObject = class(TInterfacedPersistent, IStreamPersist)
  public
    procedure SaveToStream(Stream: TStream);
    procedure LoadFromStream(Stream: TStream);
    constructor Create; virtual;
  end;

  TRttiList = class(TObjectList, IStreamPersist)
  protected
    function GetRttiObject(Index: Integer): TRttiObject;
    procedure SetRttiObject(Index: Integer; AItem: TRttiObject);
    function _AddRef: Integer; stdcall;
    function _Release: Integer; stdcall;
  public
    function QueryInterface(const IID: TGUID; out Obj):
      HResult; virtual; stdcall;
    function Add(AItem: TRttiObject): Integer;
    function Extract(AItem: TRttiObject): TRttiObject;
    function Remove(AItem: TRttiObject): Integer;
    function IndexOf(AItem: TRttiObject): Integer;
    function First: TRttiObject;
    function Last: TRttiObject;
    procedure Insert(Index: Integer; AItem: TRttiObject);
    procedure SaveToStream(Stream: TStream); dynamic;
    procedure LoadFromStream(Stream: TStream); dynamic;
    procedure SaveToFile(const FileName: TFileName); virtual;
    procedure LoadFromFile(const FileName: TFileName); virtual;
    property Items[Index: Integer]: TRttiObject read GetRttiObject write
      SetRttiObject; default;
  end;

  //    
  //       :
  // MyList.BeginUpdate;
  // try
  //   DoSomething;
  // finally
  //   MyList.EndUpdate;
  // end;
  TAsyncRttiList = class(TRttiList)
  private
    FCriticalSection: TRTLCriticalSection;
    procedure WaitReleaseCriticalSection;
  protected
    //    /.  
    // PostMessage
    procedure NotifyMainThread(NotifyType: TNotifyType); virtual; abstract;
  public
    constructor Create; virtual;
    destructor Destroy; override;
    procedure BeginUpdate;
    procedure EndUpdate;
    procedure AsyncLoadFromFile(const FileName: TFileName);
    procedure AsyncSaveToFile(const FileName: TFileName);
  end;

  TSaveLoadThread = class(TThread)
  private
    FListObj: TAsyncRttiList;
    FFileName: TFileName;
  public
    property ListObj: TAsyncRttiList read FListObj write FListObj;
    property FileName: TFileName read FFileName write FFileName;
  end;

  TSaveThread = class(TSaveLoadThread)
  protected
    procedure Execute; override;
  end;

  TLoadThread = class(TSaveLoadThread)
  protected
    procedure Execute; override;
  end;

implementation

{ TRttiList }

function TRttiList.Add(AItem: TRttiObject): Integer;
begin
  Result := inherited Add(AItem);
end;

function TRttiList.Extract(AItem: TRttiObject): TRttiObject;
begin
  Result := TRttiObject(inherited Extract(AItem));
end;

function TRttiList.First: TRttiObject;
begin
  Result := TRttiObject(inherited First);
end;

function TRttiList.GeTRttiObject(Index: Integer): TRttiObject;
begin
  Result := TRttiObject(inherited Items[Index]);
end;

function TRttiList.IndexOf(AItem: TRttiObject): Integer;
begin
  Result := inherited IndexOf(AItem);
end;

procedure TRttiList.Insert(Index: Integer; AItem: TRttiObject);
begin
  inherited Insert(Index, AItem);
end;

function TRttiList.Last: TRttiObject;
begin
  Result := TRttiObject(inherited Last);
end;

procedure TRttiList.LoadFromFile(const FileName: TFileName);
var
  Stream: TFileStream;
begin
  Stream:=TFileStream.Create(FileName,fmOpenRead);
  try
    LoadFromStream(Stream);
  finally
    Stream.Free;
  end;
end;

procedure TRttiList.LoadFromStream(Stream: TStream);
var
  Cnt,i: Integer;
begin
  Clear;
  Stream.Read(Cnt,SizeOf(Integer));
  Capacity:=Cnt;
  for i:=0 to Cnt - 1 do begin
    Add(TRttiObject.Create);
    Items[i].LoadFromStream(Stream);
  end;
end;

function TRttiList.QueryInterface(const IID: TGUID; out Obj): HResult;
const
  E_NOINTERFACE = HResult($80004002);
begin
  if GetInterface(IID, Obj) then Result := 0 else Result := E_NOINTERFACE;
end;

function TRttiList.Remove(AItem: TRttiObject): Integer;
begin
  Result := inherited Remove(AItem);
end;

procedure TRttiList.SaveToFile(const FileName: TFileName);
var
  Stream: TFileStream;
begin
  Stream:=TFileStream.Create(FileName,fmCreate);
  try
    SaveToStream(Stream);
  finally
    Stream.Free;
  end;
end;

procedure TRttiList.SaveToStream(Stream: TStream);
var
  i: Integer;
begin
  Stream.Write(Count,SizeOf(Integer));
  for i:=0 to Count - 1 do begin
    Items[i].SaveToStream(Stream);
  end;
end;

procedure TRttiList.SeTRttiObject(Index: Integer; AItem: TRttiObject);
begin
  inherited Items[Index] := AItem;
end;

function TRttiList._AddRef: Integer;
begin
  Result := -1;
end;

function TRttiList._Release: Integer;
begin
  Result := -1;
end;

{ TRttiObject }

constructor TRttiObject.Create;
begin
  inherited Create;
end;

procedure TRttiObject.LoadFromStream(Stream: TStream);
var
  TypeData: PTypeData;
  PropList: PPropList;
  Count,i: Integer;

  procedure ReadOrdProp;
  var
    Value: Integer;
  begin
    Stream.Read(Value,SizeOf(Value));
    SetOrdProp(self,PropList[i],Value);
  end;

  procedure ReadFloatProp;
  var
    Value: Extended;
  begin
    Stream.Read(Value,SizeOf(Value));
    SetFloatProp(self,PropList[i],Value);
  end;

  procedure ReadStringProp;
  var
    Value: String;
    L: Integer;
  begin
    Stream.Read(L,SizeOf(L));
    SetLength(Value,L);
    Stream.Read(PChar(Value)^,L);
    SetStrProp(self,PropList[i],Value);
  end;

  procedure ReadClassProp;
  var
    Obj: TObject;
    SaveLoader: IStreamPersist;
    IsEmpty: Boolean;
  begin
    IsEmpty:=false;
    Obj:=GetObjectProp(self,PropList[i]);
    if (Obj is TGraphic) then begin
      Stream.Read(IsEmpty,SizeOf(Boolean));
    end;
    if (not IsEmpty) then begin
      if Supports(Obj,IStreamPersist,SaveLoader) then begin
        SaveLoader.LoadFromStream(Stream);
      end;
    end;
  end;
  
begin
  TypeData:=GetTypeData(ClassInfo);
  Count:=TypeData.PropCount;
  if Count>0 then begin
    GetMem(PropList,SizeOf(PPropInfo)*Count);
    try
      GetPropInfos(ClassInfo,PropList);
      for i:=0 to Count - 1 do begin
        case PropList[i].PropType^.Kind of
          tkEnumeration, tkInteger, tkChar, tkWChar: ReadOrdProp;
          tkFloat: ReadFloatProp;
          tkString, tkLString: ReadStringProp;
          tkClass: ReadClassProp;
        end;
      end;
    finally
      FreeMem(PropList,SizeOf(PPropInfo)*Count);
    end;
  end;
end;

procedure TRttiObject.SaveToStream(Stream: TStream);
var
  TypeData: PTypeData;
  PropList: PPropList;
  Count,i: Integer;

  procedure WriteOrdProp;
  var
    Value: Integer;
  begin
    Value:=GetOrdProp(self,PropList[i]);
    Stream.Write(Value,SizeOf(Value));
  end;

  procedure WriteFloatProp;
  var
    Value: Extended;
  begin
    Value:=GetFloatProp(self,PropList[i]);
    Stream.Write(Value,SizeOf(Value));
  end;

  procedure WriteStringProp;
  var
    Value: String;
    L: Integer;
  begin
    Value:=GetStrProp(self,PropList[i]);
    L:=Length(Value);
    Stream.Write(L,SizeOf(L));
    Stream.Write(PChar(Value)^,Length(Value));
  end;

  procedure WriteClassProp;
  var
    Obj: TObject;
    SaveLoader: IStreamPersist;
    IsEmpty: Boolean;
  begin
    Obj:=GetObjectProp(self,PropList[i]);
    if (Obj is TGraphic) then begin
      IsEmpty:=TGraphic(Obj).Empty;
      Stream.Write(IsEmpty,SizeOf(Boolean));
    end;
    if Supports(Obj,IStreamPersist,SaveLoader) then begin
      SaveLoader.SaveToStream(Stream);
    end;
  end;

begin
  TypeData:=GetTypeData(ClassInfo);
  Count:=TypeData.PropCount;
  if Count>0 then begin
    GetMem(PropList,SizeOf(PPropInfo)*Count);
    try
      GetPropInfos(ClassInfo,PropList);
      for i:=0 to Count - 1 do begin
        case PropList[i].PropType^.Kind of
          tkEnumeration, tkInteger, tkChar, tkWChar: WriteOrdProp;
          tkFloat: WriteFloatProp;
          tkString, tkLString: WriteStringProp;
          tkClass: WriteClassProp;
        end;
      end;
    finally
      FreeMem(PropList,SizeOf(PPropInfo)*Count);
    end;
  end;
end;

{ TAsyncRttiList }

procedure TAsyncRttiList.AsyncLoadFromFile(const FileName: TFileName);
var
  Thread: TLoadThread;
begin
  try
    Thread:=TLoadThread.Create(true);
    Thread.FreeOnTerminate:=true;
    Thread.ListObj:=self;
    Thread.FileName:=FileName;
    Thread.Resume;
  except
    //  
  end;
end;

procedure TAsyncRttiList.AsyncSaveToFile(const FileName: TFileName);
var
  Thread: TSaveThread;
begin
  try
    Thread:=TSaveThread.Create(true);
    Thread.FreeOnTerminate:=true;
    Thread.ListObj:=self;
    Thread.FileName:=FileName;
    Thread.Resume;
  except
    //  
  end;
end;

procedure TAsyncRttiList.BeginUpdate;
begin
  EnterCriticalSection(FCriticalSection);
end;

constructor TAsyncRttiList.Create;
begin
  inherited Create;
  InitializeCriticalSection(FCriticalSection);
end;

destructor TAsyncRttiList.Destroy;
begin
  WaitReleaseCriticalSection;
  DeleteCriticalSection(FCriticalSection);
  inherited Destroy;
end;

procedure TAsyncRttiList.EndUpdate;
begin
  LeaveCriticalSection(FCriticalSection);
end;

procedure TAsyncRttiList.WaitReleaseCriticalSection;
begin
  EnterCriticalSection(FCriticalSection);
  LeaveCriticalSection(FCriticalSection);
end;

{ TSaveThread }

procedure TSaveThread.Execute;
begin
  ListObj.BeginUpdate;
  try
    ListObj.SaveToFile(FFileName);
    ListObj.NotifyMainThread(ntSaveComplete);
  finally
    ListObj.EndUpdate;
  end;
end;

{ TLoadThread }

procedure TLoadThread.Execute;
begin
  ListObj.BeginUpdate;
  try
    ListObj.LoadFromFile(FFileName);
    ListObj.NotifyMainThread(ntLoadComplete);
  finally
    ListObj.EndUpdate;
  end;
end;

end.
